from datetime import timedelta, datetime


def phrase_to_date(phrase):
    """
    Takes a typical phrase such as "3_days_ago" and returns a
    date based on that phrase.
    :phrase: str. Something like 3_days_ago or 25_days_ago etc.
    :return: '%Y-%m-%d'
    """
    try:
        date_delta = int(phrase.split('_')[0])
        return (datetime.now() - timedelta(days=date_delta)).strftime('%Y-%m-%d')
    except ValueError as err:
        raise ValueError(f'Phrasing for date: {phrase} is not valid, '
                         f'try something like: 3_days_ago') from err


def date_range(start_date, end_date, delta=timedelta(days=1)):
    """
    The range is inclusive, so both start_date and end_date will be returned.
    :start_date: The datetime object representing the first day in the range.
    :end_date: The datetime object representing the second day in the range.
    :delta: A datetime.timedelta instance, specifying the step interval. Defaults to one day.
    Yields:
        Each datetime object in the range.
    """

    if start_date == 'today':
        start_date = datetime.now().strftime('%Y-%m-%d')
    elif start_date == 'yesterday':
        start_date = (datetime.now() - timedelta(days=1)).strftime('%Y-%m-%d')
    elif '_days_ago' in start_date:
        start_date = phrase_to_date(start_date)
    else:
        raise ValueError('StartDate requires a valid input '
                         'such as "today", "yesterday" or "<int>_days_ago".')

    if end_date == 'today':
        end_date = datetime.now().strftime('%Y-%m-%d')
    elif end_date == 'yesterday':
        end_date = (datetime.now() - timedelta(days=1)).strftime('%Y-%m-%d')
    elif '_days_ago' in end_date:
        end_date = phrase_to_date(end_date)
    else:
        raise ValueError('EndDate requires a valid input '
                         'such as "today", "yesterday" or "<int>_days_ago".')

    start_date = datetime.strptime(start_date, '%Y-%m-%d')
    end_date = datetime.strptime(end_date, '%Y-%m-%d')

    current_date = start_date
    while current_date <= end_date:
        yield current_date.strftime('%Y-%m-%d')
        current_date += delta


def un_nest_keys(data, col, key_list, value_list):
    """
    A simple method that takes a set of keys and values
    from a dataset with dimensions and replaces them
    with key value pairs.

    :data: The dataset that needs to be un_nested.
    :col: The key name of the dimensions.
    :key_list: List of keys.
    :value_list: List of Values.

    returns:
        So dataset A:
            {'keys': [1, 2, 3]}
        Can become:
            {'key1': 1, 'key2': 2, 'key3': 3}
    """
    data.update(
        dict(zip(key_list, value_list))
    )
    del data[col]
    return data



"""
    CUSTOM FALCON READER CLASSES
"""
from contextlib import contextmanager
from datetime import datetime, timedelta

import os
import requests
import simplejson
import yaml
from sdc_dp_helpers.api_utilities.retry_managers import request_handler


class CustomFalconReader:
    """
        Custom Falcon Reader
    """

    def __init__(self, api_key_path, config_path=None):
        self.api_key_path = api_key_path
        self.config_path = config_path

        self.api_key = self.get_api_key()
        self.config = self.get_configuration()

        self.request_session = requests.Session()

    def get_api_key(self):
        """
            Gathers key and value pairs of client and their token.
        """
        with open(self.api_key_path, 'r') as file:
            data = yaml.safe_load(file)
            return data.get('api_key')

    def get_configuration(self):
        """
            Gathers optional and required configuration that
            defines the params for the api. This is optional.
            The configuration resembles the parameters of the
            given api method. eg:

            since: '2020-10-01T22:00:00.000Z'
            until: '2021-02-09T22:00:00.000Z'
            channels:
                - 'twitter_110592_91962761'
                - 'linkedin_110592_4840205'
            statuses:
                - 'all'
            limit: '30'
        """
        if self.config_path is not None:
            with open(self.config_path, 'r') as file:
                data = yaml.safe_load(file)
                return data
        return None

    def get_channel_ids(self):
        """
            Gather all available channel ids.
        """
        url = 'https://api.falcon.io/channels?apikey={apikey}' \
            .format(apikey=self.api_key)

        # no configs required, everything should be fetched, set limit to as high as possible
        try:
            response = requests.get(url=url, params={'limit': '999'}).json()
        except simplejson.errors.JSONDecodeError as err:
            raise EnvironmentError from err

        data_set = []
        for items in response['items']:
            data_set.append(items['id'])

        return data_set

    @request_handler(
        wait=int(os.environ.get('REQUEST_WAIT_TIME', 0.1)),
        backoff_factor=float(os.environ.get('REQUEST_BACKOFF_FACTOR', 0.01)),
        backoff_method=os.environ.get('REQUEST_BACKOFF_METHOD', 'random')
    )
    @contextmanager
    def metrics_by_channel_context(self,
                                   channel_ids,
                                   offset=0,
                                   filter_now_type='all',
                                   filter_now=True):
        """
        Get metrics by channel Id context returns a request session with the ability
        to page with offsets.
        :channel_ids: list. List of channels for the API to return metrics on.
        :offset: int. Offset number from 0 until None is returned.
        :filter_now_type: str. filter type of 'all', 'default' or 'post'
        :filter_now: bool. Flag to set date filter to dynamic, in this case yesterdays data.
                     (today often returns none)

        To use the context use 'with', expect none when no more data is available in the
        context request session:

        offset = 0
        while True:
            with reader.metrics_by_channel_context(channel_ids=channels, offset=offset) as response:
                data = response.json()

                if data is None:
                    break

            print(data)
            offset += 1

        """
        url = 'https://api.falcon.io/measure/api/v1/content/metrics?apikey={apikey}' \
            .format(apikey=self.api_key)

        params = self.config
        params['channels'] = channel_ids

        # there are three options for dynamic time filtering
        if filter_now:
            now = (datetime.now() - timedelta(1)).strftime('%Y-%m-%d')
            if filter_now_type == 'default':
                params['since'] = now
                params['until'] = now
            elif filter_now_type == 'post':
                params['postsSince'] = now
                params['postsUntil'] = now
            elif filter_now_type == 'all':
                params['since'] = now
                params['until'] = now
                params['postsSince'] = now
                params['postsUntil'] = now
            else:
                raise ValueError('{val} is not a valid filter_today_type option'.format(
                    val=filter_now_type
                ))

        params['offset'] = offset
        try:
            yield self.request_session.post(url=url, json=params)
        except Exception as err:
            raise err
